"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const thrift_parser_1 = require("./@creditkarma/thrift-parser");
const handleComments_1 = require("./handleComments");
const fs = tslib_1.__importStar(require("fs-extra"));
const utils_1 = require("../tools/utils");
async function readCode(filefullname, options, // TODO 添加option支持
includeMap) {
    const source = await fs.readFile(filefullname);
    return parser(filefullname, source.toString(), options, includeMap);
}
exports.readCode = readCode;
function parser(filefullname, source, options, // TODO 添加option支持
includeMap) {
    const ast = thrift_parser_1.parse(source);
    if (!isThritDocument(ast)) {
        throw new Error('thrift parser error:' + filefullname);
    }
    handleComments_1.handleComments(ast);
    if (options && options.annotationConfigPath) {
        if (fs.existsSync(options.annotationConfigPath)) {
            try {
                options.annotationConfig = JSON.parse(fs.readFileSync(options.annotationConfigPath).toString());
            }
            catch (e) {
                console.error(e);
            }
        }
    }
    const rtn = {
        ns: '',
        fileName: filefullname,
        includes: [],
        interfaces: [],
        enums: [],
        typeDefs: [],
        services: []
    };
    const namespaces = {};
    ast.body.forEach((ts) => {
        // namespace
        if (ts.type === thrift_parser_1.SyntaxType.NamespaceDefinition) {
            // namespace 的处理逻辑，抓一个就来了
            // TODO 优先考虑js的namespace，之后是go，再之后随便抓一个
            namespaces[ts.scope.value] = ts.name.value;
        }
        // includes
        if (ts.type === thrift_parser_1.SyntaxType.IncludeDefinition) {
            rtn.includes.push(ts.path.value);
        }
        // struct like: struct exception union --> interfaces
        if (ts.type === thrift_parser_1.SyntaxType.StructDefinition ||
            ts.type === thrift_parser_1.SyntaxType.UnionDefinition ||
            ts.type === thrift_parser_1.SyntaxType.ExceptionDefinition) {
            const name = ts.name.value;
            const aInterface = {
                name,
                properties: {},
                loc: ts.loc,
                comments: ts.comments,
                commentsAfter: ts.commentsAfter,
                commentsBefore: ts.commentsBefore
            };
            // 添加属性
            ts.fields.forEach((field) => {
                const { entity: temp, name } = handleField(field, options);
                aInterface.properties[name] = temp;
            });
            rtn.interfaces.push(aInterface);
        }
        // typedef
        if (ts.type === thrift_parser_1.SyntaxType.TypedefDefinition) {
            const aTypeDef = {
                type: '',
                alias: '',
                comments: ts.comments,
                commentsAfter: ts.commentsAfter,
                commentsBefore: ts.commentsBefore,
                loc: ts.loc
            };
            aTypeDef.alias = ts.name.value;
            aTypeDef.type = getFieldTypeString(ts.definitionType);
            rtn.typeDefs.push(aTypeDef);
        }
        // service
        if (ts.type === thrift_parser_1.SyntaxType.ServiceDefinition) {
            const aService = {
                name: ts.name.value,
                interfaces: {},
                comments: ts.comments,
                commentsAfter: ts.commentsAfter,
                commentsBefore: ts.commentsBefore,
                loc: ts.loc
            };
            ts.functions.forEach(func => {
                aService.interfaces[func.name.value] = handleFunction(func, options);
            });
            rtn.services.push(aService);
        }
        // enum
        if (ts.type === thrift_parser_1.SyntaxType.EnumDefinition) {
            rtn.enums.push(handleEnum(ts));
        }
        // const 考虑支持
        if (ts.type === thrift_parser_1.SyntaxType.ConstDefinition) {
            //
        }
    });
    const namespacesValues = Object.values(namespaces);
    rtn.ns = namespaces.js || namespaces.go;
    if (!rtn.ns && namespacesValues.length) {
        rtn.ns = namespacesValues[0];
    }
    /* istanbul ignore if */
    if (includeMap) {
        includeMap[rtn.fileName] = rtn;
    }
    return rtn;
}
exports.parser = parser;
function isThritDocument(ast) {
    return ast.type === thrift_parser_1.SyntaxType.ThriftDocument;
}
/**
 *
 * @param {ThriftDocument} ast thrift-parser输出的ast
 * @returns {RpcEntity} 返回rpc
 */
function transformAst(ast) {
    return {};
}
exports.transformAst = transformAst;
function getFieldTypeString(fieldType) {
    const ThriftType2JavascriptType = {
        [thrift_parser_1.SyntaxType.BoolKeyword]: 'boolean',
        [thrift_parser_1.SyntaxType.ByteKeyword]: 'number',
        [thrift_parser_1.SyntaxType.I16Keyword]: 'number',
        [thrift_parser_1.SyntaxType.I32Keyword]: 'number',
        [thrift_parser_1.SyntaxType.I64Keyword]: 'Int64',
        [thrift_parser_1.SyntaxType.DoubleKeyword]: 'number',
        [thrift_parser_1.SyntaxType.StringKeyword]: 'string',
        [thrift_parser_1.SyntaxType.BinaryKeyword]: 'any',
        [thrift_parser_1.SyntaxType.ListKeyword]: 'Array',
        [thrift_parser_1.SyntaxType.MapKeyword]: 'Map',
        [thrift_parser_1.SyntaxType.SetKeyword]: 'Set',
        // UPDATE 添加void
        [thrift_parser_1.SyntaxType.VoidKeyword]: 'void'
    };
    if (fieldType.type === thrift_parser_1.SyntaxType.Identifier) {
        return fieldType.value;
    }
    if (fieldType.type === thrift_parser_1.SyntaxType.SetType) {
        return `Set<${getFieldTypeString(fieldType.valueType)}>`;
    }
    if (fieldType.type === thrift_parser_1.SyntaxType.ListType) {
        return `${getFieldTypeString(fieldType.valueType)}[]`;
    }
    if (fieldType.type === thrift_parser_1.SyntaxType.MapType) {
        return `Map<${getFieldTypeString(fieldType.keyType)}, ${getFieldTypeString(fieldType.valueType)}>`;
    }
    return ThriftType2JavascriptType[fieldType.type];
}
function handleField(field, options = {}) {
    let name = field.name.value;
    const commentsBefore = field.commentsBefore || [];
    // 需要处理typedef
    const type = getFieldTypeString(field.fieldType);
    const index = field.fieldID ? field.fieldID.value : 0;
    let optional = options.useStrictMode
        ? field.requiredness !== 'required'
        : !(field.requiredness !== 'optional');
    // 考虑多种type数据的default value StringLiteral | IntConstant | DoubleConstant | BooleanLiteral | ConstMap | ConstList | Identifier
    let defaultValue;
    if (field.defaultValue !== null) {
        switch (field.defaultValue.type) {
            case thrift_parser_1.SyntaxType.StringLiteral:
                defaultValue = field.defaultValue.value;
                break;
            case thrift_parser_1.SyntaxType.IntConstant:
                defaultValue = field.defaultValue.value.value;
                break;
            case thrift_parser_1.SyntaxType.DoubleConstant:
                defaultValue = field.defaultValue.value.value;
                break;
            case thrift_parser_1.SyntaxType.BooleanLiteral:
                defaultValue = String(field.defaultValue.value);
                break;
            case thrift_parser_1.SyntaxType.ConstMap:
                // 简单的处理
                defaultValue = 'Map';
                break;
            case thrift_parser_1.SyntaxType.ConstList:
                // 简单的处理2
                defaultValue = 'List';
                break;
            case thrift_parser_1.SyntaxType.Identifier:
                defaultValue = field.defaultValue.value;
                break;
        }
    }
    if (options.useTag) {
        const annotations = field.annotations;
        const tagValueReg = /json:"([\w_\d]+).*"/;
        const tagNameReg = /(\w+).tag/;
        if (annotations) {
            const nameTag = annotations.annotations.find((annotation) => {
                let match;
                if ((match = tagNameReg.exec(annotation.name.value))) {
                    return match[1] === options.useTag;
                }
                return false;
            });
            if (nameTag) {
                const match = tagValueReg.exec(nameTag.value ? nameTag.value.value : '');
                if (match) {
                    name = match[1];
                }
            }
        }
    }
    // annotation config 优先级高于useTag
    if (options && options.annotationConfig) {
        const { fieldComment, fieldKey } = options.annotationConfig;
        if (field.annotations &&
            Array.isArray(field.annotations.annotations) &&
            (Array.isArray(fieldComment) || fieldKey)) {
            let comment = '';
            field.annotations.annotations.forEach(annotation => {
                if (Array.isArray(fieldComment)) {
                    if (fieldComment.indexOf(annotation.name.value) > -1) {
                        comment += `@${annotation.name.value}:${annotation.value.value}    `;
                    }
                }
                if (fieldKey) {
                    if (annotation.name.value === fieldKey) {
                        name = annotation.value.value;
                    }
                }
            });
            commentsBefore.push({
                type: thrift_parser_1.SyntaxType.CommentLine,
                value: comment,
                loc: field.loc
            });
        }
    }
    if (!utils_1.isUndefined(defaultValue)) {
        // 如果有默认值，不需要指定 optional
        optional = true;
        let value = defaultValue;
        if (defaultValue === '') {
            value = '""';
        }
        commentsBefore.push({
            type: thrift_parser_1.SyntaxType.CommentLine,
            value: `@default: ${value}`,
            loc: field.loc
        });
    }
    else {
        defaultValue = '';
    }
    return {
        entity: {
            type,
            index,
            optional,
            defaultValue,
            comments: field.comments,
            commentsBefore,
            commentsAfter: field.commentsAfter,
            loc: field.loc
        },
        name
    };
}
/*
  returnType: string;
  inputParams: Array<{
    type: string;
    index: number;
    name: string;
  }>;
  comment: string;
 */
function handleFunction(func, options) {
    const returnType = getFieldTypeString(func.returnType);
    const inputParams = func.fields.map((field) => {
        const { entity: temp, name } = handleField(field, options);
        return {
            type: temp.type,
            index: temp.index,
            name
        };
    });
    let comment = '';
    if (options && options.annotationConfig) {
        // 根据annotation生成config
        const { functionMethod, functionUri } = options.annotationConfig;
        if (functionMethod || functionUri) {
            if (func.annotations && Array.isArray(func.annotations.annotations)) {
                func.annotations.annotations.forEach(annotation => {
                    if (functionMethod) {
                        if (annotation.name.value === functionMethod) {
                            comment += `@method: ${annotation.value.value}    `;
                        }
                    }
                    if (functionUri) {
                        if (annotation.name.value === functionUri) {
                            comment += `@uri: ${annotation.value.value}    `;
                        }
                    }
                });
            }
        }
    }
    const commentsBefore = func.commentsBefore || [];
    if (comment) {
        commentsBefore.push({
            loc: func.loc,
            value: comment,
            type: thrift_parser_1.SyntaxType.CommentLine
        });
    }
    return {
        returnType,
        inputParams,
        comments: [],
        loc: func.loc,
        commentsBefore,
        commentsAfter: func.commentsAfter
    };
}
function handleEnum(e, options) {
    const name = e.name.value;
    const properties = {};
    let currentValue = 0;
    e.members.forEach(member => {
        /**
         * https://wiki.apache.org/thrift/Tutorial
         * You can define enums, which are just 32 bit integers. Values are optional
         * and start at 1 if not supplied, C style again.
         *              ^ ThriftIDL page says "If no constant value is supplied,
         *   the value is either 0 for the first element, or one greater than the
         *   preceding value for any subsequent element" so I'm guessing that's a bug.
         *   PS: http://enel.ucalgary.ca/People/Norman/enel315_winter1997/enum_types/ states that if values are not supplied, they start at 0 and not 1.
         */
        let value;
        if (member.initializer) {
            value = parseInt(member.initializer.value.value, 10);
        }
        else {
            value = currentValue++;
        }
        properties[member.name.value] = {
            value,
            loc: member.loc,
            comments: member.comments,
            commentsBefore: member.commentsBefore,
            commentsAfter: member.commentsAfter
        };
    });
    return {
        name,
        properties,
        loc: e.loc,
        comments: e.comments,
        commentsBefore: e.commentsBefore,
        commentsAfter: e.commentsAfter
    };
}
